;****************************************************************
;                     McCoi II- Ketten ler                     ;
;                                                               ;
;          AVR 90S2313, 4 MHz, externer Oszillator              ;
;                                                               ;
;      (c) 2001 MDvP                   mailto:MDvP@gmx.de		;
;      nderungen fr 90S2313: 12/2001 T. Ostermann             ;
;      Version 1.04   28.07.2002                                ;
;                                                               ;
;                                                               ;
;      Version 1.10   10.08.2003 Steffen Koberitz               ;
;      - Radkonstante auf 333m erhht                           ;
;      - Bereich von 2650m bis 7650m erweitert                  ;
;      - Kontrolle Regenfunktion auf PD1 (bisher frei) gelegt   ;
;      - Pumpimpuls-Kontrolle auf PD0 (bisher frei) gelegt      ;
;        damit kann Kontrollschaltung ohne Relais angesteuert!  ;
;
; Version 1.20 19.09.2004 von Steffen Koberitz
; - Speziell fr den neuen vergoldetetn Regensensor wurde
;   die Abfrage nach dem Regenstatus angepat. Sonst keine
;   nderungen
;
; - Version 1.21 
;   vom 13.02.2005 by Steffen Koberitz
; - Bug beseitigt: Abfrage fr Regenmodus war vertauscht,
;   nun korrekte Abfrage
;   Lieferungen ab inkl. Montag 14.02.05 haben die neue, 
;   korrekte Version!
;
;****************************************************************
; Version 1.22 
; - vom 22.12.2005 von Steffen Koberitz
;   Die Externe Regen-LED wird an PD1 angeschlossen
;   Der Minus-Pol der LED mu an PD1 
;   und der Plus-Pol an VCC (+5Volt) -> Vorwiderstand nicht vergessen!
;   Damit glimmt die LED nicht mehr wegen Leckstrom
;   Die LED wird auch im Sleepmodus abgeschaltet! :-)
; - Pumpensteuerzeit von bisher 50ms auf 150ms verlngert
;   damit sollte es keine Viskosittsprobleme mehr geben
;   ACHTUNG: im Testmodus weiterhin 50ms, im Betriebsmodus 150ms!!!
;
; - 14.02.2006 von Steffen Koberitz
;   BugFix: Externe Regen-LED wurde falsch angesteuert 
;	    (leuchtet bei Trockenheit)
;   Nun korrekt: Externe Regen-LED leuchtet bei Nsse!
;
; - 15.04.2006 von Steffen Koberitz
;   Anzahl der Pumpimpulse ber "Cross-Schalter" an PD0 gegen Masse
;   whlbar: blicherweise nur ein Impuls. Im Cross-Mode 8 Impulse
;   BETA!!!
;   Pumproutine wurde hierzu aus der Interruptroutine ausgegliedert!
; - 23.12.2006 Peter Bttner	
;   Bladefunktion eingefgt fr BMW F650 GS Tachosignalabgriff
;
; - 06.01.2007 Steffen Koberitz
;   Umbenennung in "STK-T-03", damit die Nomenklatur beibehalten wird.
;   Lschen der alten Kommentare und Update auf McCoi-Platine 2.5.
;   LED fr Crossmode auf PA0 gelegt: Schliee LED an VCC (PA0 als GND)
;   -> ist der Crossmode (8 Pumpimpulse) aktiv, leuchtet die LED.
;   die Konstante "Tachopulse" ist bei Reed-Betrieb auf 1 zu setzen!
;   Fr andere Bikes kann sie angepat werden (z.B. bei Bandit auf 12).
;
; - 11.06.07, Martin Kirsch
;	Changelog kann im McCoi-Forum nachgelesen werden:
;	http://www.mccoi-forum.de/viewtopic.php?t=724
;	Stichworte:
;	- Extralen nach Regen
;	- Notlaufmodus
;	- Geschwindigkeitsprogression
;	- Regenmodusverkrzung abhngig von Hex-Schalterstellung
;	- Programm um- und aufgerumt
;	- Sleepmode im Ausnahmezweig (geschalteter Reed) ohne Controller-Reset
;	- Zustzliche Konstanten, um Programm auf eigene Bedrfnisse anzupassen
;
; - 22.08.07, Martin Kirsch
;	Teachmodus hinzugefgt:
;	- McCoi kann Anzahl Tachopulse und Radkonstante lernen, kein Programmieraufwand mehr ntig
;	Wie funktionierts?
;	- McCoi anschlieen
;	- Hexschalter auf eine beliebige Stellung drehen, AUSSER Hex0
;	- 2 Sekunden den Taster auf der Platine drcken -> Reed-LED auf der Platine leuchtet auf
;	  -> Teachmodus Tachopulse aktiv
;	- jetzt 10mal das Rad drehen, welches die Impulse liefert
;	- wieder 2 Sekunden den Taster drcken -> Reed-LED erlischt, Regen-LED leuchtet auf
;	  -> gelernte Tachokonstante wird im EEPROM gespeichert und ab sofort fr
;		 die Ermittlung der lintervalle genutzt.
;	  -> Teachmodus Radkonstante aktiv
;	- jetzt mglichst exakt 1000 Meter fahren
;	- nochmal 2 Sekunden den Taster drcken -> Regen-LED erlischt
;	  -> Teachmodus deaktiviert, gelernte Radkonstante wird im EEPROM gespeichert und ab sofort fr
;		 die Ermittlung der lintervalle genutzt.
;	Details: http://www.mccoi-forum.de/viewtopic.php?t=870
;
; - 28.10.08, Martin Kirsch
;	es hat sich viel getan
;	- Geschwindigkeitsprogression per Tabelle frei einstellbar
;	- nderbare Programmierung des Hex-Schalters
;	- verzgerte Regenumschaltung, um Flackern von RegenLED auszuschalten
;	- eingestellte Parameter werden im EEPROM abgelegt
; 	- ein Windows-Tool zum Konfigurieren und Programmieren des McCoi:
;	Konfigtool: http://mccoi-forum.de/viewtopic.php?t=1212
;	Konfigtool 1.3: http://mccoi-forum.de/viewtopic.php?t=1321
;	minor Update (Sofortpumpen bei Crossmode) fr 1.3: http://mccoi-forum.de/viewtopic.php?t=1328
;
; - 07.11.08, Martin Kirsch
;	- Bugfix, ein Tippfehler (vergessener Unterstrich) korrigiert
;
; - 09.02.09, Martin Kirsch
;	- Bugfix, Sleepmode war auch bei per Software deaktiviertem Notlaufmodus
;	  vom Status des Znd(+)-Pins abhngig
;	-> teilweise kein Sleep, wenn die Ltbrcke nicht gesetzt wurde, obwohl der
;		Notlauf in der Software deaktiviert war
; - 14.02.09, Martin Kirsch
;   - Bugfix, Nachlen nach Regen war bei Intervall > 444 Meter nicht mglich
;
; - 18.08.09, Martin Kirsch
;   - alternativ zum Cross-Schalter auch Taster nutzbar
;   - nderung im EEPROM-Handling (Speicherstelle 0 wird nichtmehr angefat,
;	10 ms Delay vor jedem Schreibvorgang)
;   - mglichkeit, Pumpimpulse im EEPROM zu zhlen, work-in-progress
;
;****************************************************************
;****************************************************************

.include "tiny2313.inc"		; Definitionen fr den tiny2313

.CSEG								; Code Segment

; -----Interrupts
.ORG 	RSTaddr
		rjmp	Reset				; Program Start
.ORG 	INT0addr
		rjmp	Reed				; Reed Impuls
.ORG	OC1Aaddr
		rjmp	timer1_comp_match	; Timer/Counter1 Compare Match A     
.ORG	OVF1addr
		reti						; Timer1 Overflow Handler
.ORG	OVF0addr
		reti						; Timer0 Overflow Handler
.ORG	ACIaddr
		reti
.ORG	OC0Aaddr
		rjmp	timer0_comp_match	; Timer/Counter0 Compare Match A
.ORG    OC0Baddr
		rjmp	timer0_comp_match_b	; Timer/Counter0 Compare Match B
.ORG	WDTaddr
		reti						; Watchdog Interrupt

.db $28,$63,$29,$20										; (c)
.db $32,$30,$31,$33, $20, $20									; 2013
.db $53, $74, $65, $66, $66, $65, $6e, $20, $4b, $6f, $62, $65, $72, $69, $74, $7a, $20, $20 	; Steffen Koberitz
.db $77, $77, $77, $2e, $6d, $63, $63, $6f, $69, $2e, $64, $65, $20, $20			; www.mccoi.de


;############################################################################
;##                                                                        ##
;##                   Anfang Konfigurationsabschnitt                       ##
;##                                                                        ##
;############################################################################

;>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
; Radkonstanten fr 333m     #
.equ	Tachopulse	=1	; Anzahl Impulse pro Raddrehung
.equ	Rad		=168	; wie oft dreht sich das Rad whrend 333 m (quasi 333 / Radumfang)
					; 168 fr 180/55-17 (Radumfang ca. 1,98 m)
.equ	ReedDelay	=15	; Delay zum Ausfiltern von Doppelimpulsen
				; des Reedsensors, Angabe in Millisekunden
				; maximal sind 55 ms mglich, Werte ber 25-30 ms wenig sinnvoll
				; eine Radumdrehung bei 300 km/h und 2 Meter Radumfang dauert ca. 24 ms

.equ	Delay_automatic	=2	; Wert fr die automatische Umschaltung Delay on/off
.equ	Intervall 	=333	; Arbeitsintervall des McCoi, Unteschied zwischen 2 benachbarten Hex-Schalterstellungen
.equ	Hex_Offset 	=14	; krzester Pumpabstand (Hex1): (Hex_Offset + 1) * Intervall
#define Force_ReedDelay_On (0)	; ReedDelay immer an
#define Force_ReedDelay_Off (0) ; ReedDelay immer aus
				; stehen beide auf 0 -> Delay wird nach Anzahl der Tachopulse automatisch an oder aus
				; gesetzt, siehe -> Delay_automatic
;<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

;>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
;Einstellungen frs "Nach dem Regen Extralen"
;fr die Strecken bitte nur Werte >0 verwenden
.equ	RStrecke	=5	; Mindeststrecke Regen in km
.equ	TStrecke	=1	; Mindeststrecke Trocken in km
.equ	Pumpstoesse 	=7	; nach mindestens $RStrecke Regen und anschlieend $TStrecke Trockenheit
				; werden $Pumpstoesse ausgelst, jeweils einer alle 333 Meter
				; ".equ Pumpstoesse = 0" setzen um Extralen nach Regen zu deaktivieren
#define Enable_ExtraOelen (1) 	; Nachlen an/ausschalten, "1" = angeschaltet
;<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

;>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
; Einstellungen fr den Notlaufmodus
; Zeitangaben in 15-Sekunden-Intervallen, d.h. eine Zeit von 1 Minute entspricht einem Wert von 4
; Maximalwert 255 nicht berschreiten und bitte nur Werte >0 verwenden
; NoSignal kann nicht kleiner sein als "NotIntervall"
; Update, 29.04.07:
; NotIntervall und NoSignal knnen jetzt auch dynamisch (abhngig von Hex-Schalterstellung)
; eingestellt werden. Dazu "NotIntervall" auf 0 setzen.
; Das lintervall entspricht dann ungefhr der Zeit, die man bei 100 km/h fr die
; am Hexschalter eingestellte Strecke bentigt. Dabei wird auch der Regensensor bercksichtigt.
; Bei dynamischer Einstellung wird dann bei "NoSignal" auch keine feste Zeit mehr angegeben,
; sondern ein Vielfaches von Notintervall. NoSignal mu mindestens 1 betragen. Also z.B.
; .equ NoSignal		=4
; .equ NotIntervall	=0
; -> die Lnge von "Notintervall" wird dynamisch, abhngig vom Hex-Schalter _und_ Regensensor
; berechnet, und die berechnete Zeit mu 4mal verstreichen (man verliert also bei normalem Tempo
; ca. 3 Pumpste, beim 4. Mal wird schon gelt), bevor der Controller in den Notlaufmodus geht
.equ	NoSignal	=40	; Zeit mit Zndung an und ohne Reed-Signal
.equ	NotIntervall	=8	; Pumpintervall im Notbetrieb
#define Enable_Notlauf (1)	; Notlauf per Programm aktivieren/deaktivieren, "0"= Notlauf immer aus
#define	Enable_BlinkNotlauf (1)	; auf 0 setzen, um das Blinken der Regen-LED im Notlauf zu deaktivieren
#define Enable_Notlauf_Dynamisch (1) ; Notlauf dynamisch (Zeitsteuerung abhngig von Hexschalter) an oder aus
.equ	Sleepmode	=4	; Zeit bis Sleepmode, ebenfalls vielfache von 15 Sekunden (wie bei "NoSignal" und "NotIntervall")
				; sollte sinnvollerweise kleiner als "NoSignal" sein, sonst geht der McCoi in den Notlauf,
				; anstatt einzuschlafen
;<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<



;>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
; Einstellungen Pumpe:
; - Anzahl Pumpimpulse (Normal, Regen, Cross, max. Pumpste im Testmode)
; - Dauer der Pumpimpulse + Pause im Normal- und Testmode
.equ PumpCross		=8	; Anzahl Pumpstte im Crossmode
.equ WegCross		=60	; Angabe der Wegstrecke im CrossMode, x% der normalen Wegstrecke
				; WegCross=50 -> im Crossmode ist Wegstrecke nur 50% der normalen Wegstrecke
.equ PumpNormal		=1	; im Normalmode 1 Pumpsto

.equ WegRegen		=50	; Angabe der Wegstrecke im RegenMode, x% der normalen Wegstrecke
.equ PumpRegen		=2	; Pumpste im Regen
.equ RegenDelay		=3	; mindestzeit Regen vor Umschalten in Sekunden

#define Add_Rain (0)		; 1: Pumpste Regen zustzlich zu den normalen ausfhren
							; 0: Pumpste Regen anstatt der normalen ausfhren	

#define Cross_SofortPumpen (1)  ; 1: ein Einschalten des Crossmode sofort pumpen und dann mit regulrem Crossintervall weitermachen
				; 0: erst bei abgelaufenem Intervall pumpen

#define CrossMode_Taster (0)	; 1: ein Taster wird zum schalten des Crossmode verwendet, 0: mit Schalter wie bisher.

.equ PumpTest		=50	; maximale Anzahl Pumpste im Testmode

.equ PumpOnTest		=50	; Einschaltzeit Pumpe im Testmodus (Hex0), max 255, Angabe in Millisekunden
.equ PumpOffTest	=50	; Ausschaltzeit Pumpe im Testmodus (Hex0), max 255, Angabe in Millisekunden

.equ PumpOn		=200	; Einschaltzeit Pumpe im Betriebsmodus, max 255, Angabe in Millisekunden
.equ PumpOff		=200	; Ausschaltzeit Pumpe im Betriebsmodus, max 255, Angabe in Millisekunden
;<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<


;>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
; Einstellungen geschwindigkeitsabhngiges len
;
; Geschwindigkeiten (0 bis maximal 255 km/h) in die Tabelle eintragen
; und dazu einen Wert, wieviel Prozent die Streckenlnge bei dieser Geschwindigkeit betragen soll
; Prozentwert sollte blicherweise zwischen 0 (Pumpintervalle auf 0% verkrzt) und 100 (keine Verkrzung) liegen.
; Werte ber 100 (Verlngerung der Pumpintervalle, z.B. bei niedrigen Geschwindigkeiten) sind auch mglich,
; ABER: Radkonstante * Prozentwert kann nicht grer als 255 werden.
; bei ca. 150% Verlngerung ist fr die bliche Radkonstante von 168 also Schlu:
; 168 * 150% = 252
;
; eine gltige Tabellenzeile besteht aus
;
; .db	Geschwindigkeit, Prozentwert
;
; wenn dann schneller als "Geschwindigkeit" gefahren wird, betrgt die Streckenlnge zwischen
; den Pumpsten nur noch den entsprechenden Prozentwert der normalen Streckenlnge
;
; Bsp:
; .db	130, 75
;
; Hexschalter steht auf 7 -> normal 5 km zwischen 2 Pumpsten
;
; Geschwindigkeit > 130 km/h -> nur noch 75% von 5 km = 3,75 km zwischen 2 Pumpsten

; Die Tabelle kann quasi beliebig erweitert werden (solange noch Speicherplatz vorhanden,
; aktuell ist noch fr ber 200 Eintrge Platz), zu beachten aber: die Geschwindigkeiten mssen
; der Gre nach sortiert sein, damits ordentlich funktioniert
; fr Geschwindigkeiten, die unterhalb des kleinsten in der Tabelle angegeben Geschwindigkeitswertes liegen,
; erfolgt keine Wegvernderung

#define Enable_SpeedProgression (1)	; auf 1 setzen, um geschwindigkeitsabhngiges len zu aktivieren
Speedprogression:
;	Geschwindigkeit		%wert Streckenlnge
.db 100, 95
.db 140, 90
.db 180, 85
.db 210, 80
.db 240, 75
end_speedprogression:	; diese Zeile nicht anfassen, ist wichtig fr das Erkennen des Tabellenendes
;<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<



;>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
;Einstellungen fr diverse LED-Blinkereien
; jeweils auf 1 setzen, um die Funktion zu aktivieren
#define Enable_LED_Test (1)		; alle LEDs bei Einschalten bzw. Aufwachen aus dem 
					; Sleepmode mal aufleuchten lassen
					; Regen/Cross-LED blinken 5mal
					; Kontroll-LED wird durch einen sehr kurzen (0,06 ms) Pumpimpuls eingeschaltet
					; der Pumpimpuls ist kurz genug, da dabei kein l fliet
#define	Enable_BlinkRain_for_Cross (0)	; Crossmode durch blinkende RegenLED anzeigen lassen

#define	Enable_Blink2sek_CrossLED (1)	; CrossLED alle 2 Sekunden kurz aufleuchten lassen
#define	Enable_Blink15sek_CrossLED (0)	; CrossLED alle 15 Sekunden kurz aufleuchten lassen

#define	Enable_Blink2sek_extRegenLED (0); externe RegenLED alle 2 Sekunden kurz aufleuchten lassen
#define	Enable_Blink15sek_extRegenLED (0); externe RegenLED alle 15 Sekunden kurz aufleuchten lassen
;<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<

;############################################################################
;##                                                                        ##
;##                    Ende Konfigurationsabschnitt                        ##
;##                                                                        ##
;############################################################################



.equ	Cross_SW	=PD0	; PD0 gegen Masse = Crossmode aktiv
.equ	ExternRegen	=PD1	; LED extern
.equ	ReedSW		=PD2	; (INT0)
.equ	Taste		=PD3	; 
.equ	ReedLED		=PD4	; LED auf Platine
.equ	RegenLED	=PD5	; LED auf Platine
.equ	Cross_LED	=PA0	; LED extern
.equ	Zuendung	=PA1	; Signal von Zndung+

.equ	EntprellTimer	=4*Reeddelay

.equ	CntTicksHigh=0xE4	; Gesamt 0xE4E2 = 58594 Counterticks
.equ	CntTicksLow	=0xE2	; entspricht ca. 15 Sekunden, bis zu 10% Abweichung aufgrund
							; Ungenauigkeit des Quarzes sind mglich

.equ	FlagRegister=GPIOR0 ; GPIOR0 als Register fr verschiedene Zustandsflags,
							; umbenannt um den Code leichter lesbar zu machen
.equ	Notbetrieb	=4		; Bit 4 in Flagregister fr Notbetrieb genutzt
.equ	Learn_Tacho =6		; Bit 6 in Flagregister fr Learnmode-Tachopulse genutzt
.equ	Learn_Rad	=7		; Bit 7 in Flagregister fr Learnmode-Radkonstante genutzt
.equ	Regen		=0		; Bit 0 in Flagregister fr Regen/kein Regen genutzt
.equ	Cross		=1		; Bit 1 in Flagregister fr Crossmode genutzt

.equ	RGrenze		=RStrecke*3	
.equ	TGrenze		=RGrenze+TStrecke*3
.equ	PumpMax		=TGrenze+Pumpstoesse-1
							; Konstanten fr den Nachlmodus
.equ	halbes_intervall	=(intervall+1)/2 ; Intervall bis 500 m mglich, Tiny kann aber nur 8 bit (max 255)
							; daher hier der Hilfswert fr die Verarbeitung im Programm
.equ	cntRegenGrenze	= (RegenDelay*15)+Regendelay ; Zhler fr die verzgerte Regen/Trockenumschaltung, verhindern
							; "Flackern" des Regenmodus

.equ	CntRad_prog =0x0060		; Speicherstelle Radumdrehung fr Geschwindigkeitsprogression
.equ	CntRad_rain	=0x0061		; Speicherstelle Radumdrehungen fr das Nachlen nach Regen
.equ	s_Tachopulse=0x0062		; Speicherstelle Tachopulse: am Programmstart aus EEPROM initialisiert
.equ	s_Rad		=0x0063		; Speicherstelle Radkonstante

.equ	RSchwelle	=0x0064			; Speicherstelle fr Schwellwert berwiegend Regen
.equ	TSchwelle	=0x0065			; Speicherstelle fr Schwellwert berwiegend Trocken

.equ	akt_speicherstelle_learnmode = 0x0066 ; Zeiger auf die aktuell zu nutzende SRAM-Addresse im Learnmode
.equ	s_learn_tacho_start	= 0x0067	; Start Speicherplatz fr Learnmode Tachopulse
.equ	s_learn_tacho_eeprom= 0x006b	; enthlt den nach dem Lernen zu speichernden Wert
.equ	s_learn_tacho_end	= 0x0070	; Ende Speicherplatz Learnmode Tachopulse
		; 10 Speicherpltze -> 10mal Rad drehen
.equ	s_learn_rad_start	= 0x0071	; Start Speicherplatz fr Learnmode Radkonstante
.equ	s_learn_rad_eeprom	= 0x0072	; enthlt den nach dem Lernen zu speichernden Wert
.equ	s_learn_rad_end		= 0x0073	; Ende Speicherplatz fr Learnmode Radkonstante
		; 3 Speicherpltze -> 3mal 333 Meter = 1000 Meter fahren
; Speicherbelegung (SRAM) insgesamt:
; 0x0060 -> CntRad_prog -> Zhler fr Radumdrehung in den letzten 2 Sekunden,
;			um die Geschwindigkeit berechnen zu knnen
; 0x0062 -> Tachopulse
; 0x0063 -> Radkonstante

; 0x0064 -> RSchwelle (berechnet)
; 0x0065 -> TSchwelle (berechnet)

; 0x0066 ->	Zeiger auf die aktuell zu nutzende SRAM-Speicherstelle im Learnmode
; 0x0067 -> Start Speicherplatz fr Learnmode-Tachopulse, insgesamt 10 Speicherstellen -> 10 Radumdrehungen
; 0x006b ->	enthlt den ins EEPROM zu bernehmenden Wert fr die Tachopulse
; 0x0070 -> Ende Speicherplatz fr Learnmode-Tachopulse
; 0x0071 -> Start Speicherplatz Learnmode Rad
; 0x0072 ->	enthlt den ins EEPROM zu bernehmenden Wert fr die Radkonstante
; 0x0073 -> Ende Speicherplatz fr Learnmode Rad

;-----low Register
;.def	Time1		=r1
;.def	Time2		=r2
.def	cntRegenDelay=r2		; Zhler fr die Umsetzung des Regendelays
.def	CntRad		=r3			; Zhler fr je 333 m
.def	CntWert		=r4			; Zhler fr x*333 m
.def	cPump		=r5			; Zhler Pumpenpulse beim Vorpumpen
.def	CntTacho	=r6 		; Zhler fr Tachopulse
.def	Wert		=r7			; Vergleichswert
.def	Cr_cnt		=r8			; Zhl-Wert fr Crossmode
.def	NotIntervallReg=r10		; Register fr Lnge von NotIntervall
.def	NoSignalReg =r11		; Register fr Lnge von NoSignal
.def	Cnt2Sek		=r12		; Zhlregister, um das 2-Sekunden-Intervall zu realisieren
.def	Timer1Cnt	=r13		; Zhler fr Timer1-CompMatches = 15sekunden-Zhler
;-----high Register
.def	c0			=r16		; Zhler delay-Schleife
.def	c1			=r17		; Zhler innere delay-Schleife
.def	c2			=r18		; Zhler Frequenz-Schleife
.def	temp		=r19		; Temporr-Register
.def	temp2		=r20		; zweites Temp-Register
.def	Cr_soll		=r21		; Soll-Wert fr Crossmode
.def	Regenanteil	=r22		; bestimmt die Regenmenge der letzten 333m
								; (Anteil der Reed Impulse mit geschaltetem Regensensor)
								; (25%, 75% als Schwellenwert fr nderung)
.def	Regenmerker	=r23		; Zustandsmerker fr Regensensor der letzten 333m-1666m
								; 0 - Regulrer Trockenzustand
								; 1 - letzten  333m berwiegend nass
								; 2 - letzten  666m berwiegend nass
								; ...
								; RGrenze - letzten RStreck km berwiegend nass
								; RGrenze+1 -  333m berwiegend trocken nach ausreichend nass
								; RGrenze+2 -  666m berwiegend trocken nach ausreichend nass
								; ...
								; TGrenze - nach RStrecke km Regen jetzt fr TStrecke km berwiegend trocken, len!
								; TGrenze+1 - len!
								; ...
								; PumpMax Anzahl Pumpstoesse erreicht, letztes Mal len, danach Merker wieder auf 0

.def	faktor_1	=r24		; Faktor 1
.def	quot_l		=r24		; lowbyte Ergebnis der Division
.def	faktor_2	=r25		; Faktor 2
.def	quot_h		=r25		; highbyte Ergebnis der Division
.def	mul_help	=r28		; multiplikation Hilfsregister
.def	divisor		=r28		; Divisor
.def	prod_l		=r29		; lowbyte Ergebnis der Multiplikation
.def	div_l		=r29		; lowbyte des Dividends 
.def	prod_h		=r30		; highbyte Ergebnis der Multiplikation
.def	div_h		=r30		; highbyte des Dividends 
.def	div_help	=r31		; division Hilfsregister



;XL (R26) und XH (R27) bitte NICHT umdefinieren bzw. nur mit Vorsicht verwenden,
;da die regelmig fr SRAM-Addressierung genutzt werden
;genaugenommen XH am besten garnicht anfassen, da es nur bei Programmstart
;einmal initialisiert wird und dann dauerhaft auf diesem Wert bleiben soll,
;um es nicht bei jedem Speicherzugriff neu setzen zu mssen.



;********************************************************************************************
.macro	delay
		push	c0
		ldi		C0,@0		
		rcall	do_delay
		pop		c0
.endmacro

do_delay:					; Warteschleife, x mal 1ms 
		push	c1
		push	c2		
WGLOOP0: 
		ldi		C1,$1F
WGLOOP1: 
		ldi		C2,$2A
WGLOOP2:
		dec		C2
		brne	WGLOOP2
		dec		C1
		brne	WGLOOP1
		dec		C0
		brne	WGLOOP0
		wdr
		pop		c2
		pop		c1
		ret
; delay-macro deutlich gekrzt, um Programmspeicher zu sparen
; macros werden jedesmal komplett an der verwendeten Stelle im Code eingefgt.
; per rcall aufgerufene Unterprogramm stehen nur einmal im Code und werden von
; verschiedenen Ausgangspunken aus angesprungen, daher
; eigentliche Delay-Schleifen in ein Unterprogramm ausgelagert, nur ein rudimentres
; delay-macro behalten, um den bequemen Aufruf weiterhin zu ermglichen


;********************************************************************************************
; fr die Generierung der Umschaltfrequenz des Regensensors bengtigt
.macro	freq						; Warteschleife 0,25 ms
		ldi 	c2,250
f0:		nop
		dec 	c2
		brne 	f0
		wdr
.endmacro



;********************************************************************************************
;**** Unterprogramm LED-Test ****************************************************************
;********************************************************************************************
LED_Test:							; LEDs nach Startup/Wakeup blinken lassen, um Funktion zu prfen/signalisieren
		push	temp

		sbi		PORTD,6				; Pumpe an, um die Kontrollschaltung zu triggern
		ldi 	c2,80		
LED_Test_triggerdelay:				; sehr kurze Verzgerung, damit die Kontrollschaltung Zeit hat zu reagieren
		dec 	c2
		brne 	LED_Test_triggerdelay
		cbi		PORTD,6				; Pumpe aus

		ldi		temp, 5				; LEDs 5mal blinken lassen
LED_Test_blink:		
		wdr
		cbi		PortA, Cross_LED
		cbi		PortD, Externregen	; LEDs an
		delay	200					; warten
		sbi		PortA, Cross_LED
		sbi		PortD, ExternRegen	; LEDs aus
		delay	200					; warten
		dec		temp				; Zhler um 1 verringern
		brne	LED_Test_blink		; Zhler <>0? Weiterblinken

		pop		temp
		ret


;********************************************************************************************
;************* Blink-Unterprogramm fr externe RegenLED, Anzeige Notlauf ********************
;********************************************************************************************
blink_Regen:
		push	temp
		ldi		temp,10				; LED 10mal blinken lassen
Blink_Regen_Start:
		cbi		PortD,ExternRegen	; RegenLED an
		delay	200					; warten
		sbi		PortD,ExternRegen	; RegenLED aus
		delay	200					; warten
		dec		temp				; Zhler um 1 verringern
		brne	Blink_Regen_Start	; <>0? Weiterblinken

		pop		temp
		ret


;********************************************************************************************
;************** Blink-Unterprogramm fr externe RegenLED, Anzeige Crossmode *****************
;********************************************************************************************
blink_CrossMode:
		cbi		PortD,ExternRegen		; RegenLED an
		delay	250
		delay	250						; 2 * 250 ms warten
		sbi		PortD,ExternRegen		; RegenLED aus
		sbic	Flagregister, Regen		; Regen?
		rjmp	blink_crossmode_rain	; dann nur verkrzte Aus-Zeit der LED
		delay	200
		delay	200						; 2 * 200 ms warten
blink_crossmode_rain:
		delay	100						; und nochmal 100 ms warten
		; im trockenen also 2*250 ms LED an, danach 2*200+100 ms LED aus -> 500 ms an, 500 ms aus
		; im nassen: 2*250 ms LED an, danach 100 ms LED aus -> 500 ms an, 100 ms aus
		ret


;********************************************************************************************
;***** Unterprogramm zur Berechnung der Schwellwerte fr Extralen nach Regen ***************
;********************************************************************************************
berechne_schwellwerte:
		push	temp
		push	temp2

		ldi		temp, 2				; radkonstante aus EEPROM laden
		rcall	read_eeprom
		mov		faktor_1, temp
		; Nachlen arbeitet nachwievor mit dem 333-Meter-Intervall, daher hier Anpassung an
		; mglicherweise gendertes Intervall des normalen lens
		; Radkonstante frs Nachlen auf 333 meter normieren:
		; Rad * 333 / Intervall (bzw 167 und intervall/2, da ja zahlen grer 255 nicht mglich sind)
		ldi		faktor_2, 167
		rcall 	multipliziere
		ldi		divisor, halbes_intervall
		rcall	dividiere

		sts		cntrad_rain, quot_l	; Umdrehungszhler fr Nachlen initialisieren, es wird
									; auf 0 heruntergezhlt
		mov		temp2, quot_l		; berechnete Umdrehungszahl auch in temp2 laden
		lsr		temp2
		lsr		temp2				; temp2 enthlt jetzt Umdrehung * 0,25
		sub		quot_l, temp2		; quot_l = Umdrehung * 0,75

		sts		RSchwelle, quot_l	; temp (RSchwelle) im SRAM ablegen
		sts		TSchwelle, temp2	; temp2 (TSchwelle) im SRAM ablegen

		pop		temp2
		pop		temp
		ret


;*************** Unterprogramm Multiplikation ********************
;*** geklaut von: http://www.avr-asm-tutorial.net/avr_de/rechnen/multiplikation.html
;*** berechnet aus zwei 8-Bit-Werten (Register "faktor_1" und "faktor_2"
;*** einen 16-Bit-Wert (Register prod_h, prod_l)
;*** Schade, da der Tiny2313 keinen Multiplikator hat, wrde das etwas vereinfachen ;-)
multipliziere:
		clr		mul_help	; Hilfsregister leeren
		clr		prod_l		; Ergebnisregister leeren
		clr		prod_h
mult_a:
		clc
		ror		faktor_2
		brcc	mult_b
		add		prod_l, faktor_1
		adc		prod_h, mul_help
mult_b:	
		clc
		rol		faktor_1
		rol		mul_help
		tst		faktor_2
		brne	mult_a
		ret

;*************** Unterprogramm Division ********************
;*** geklaut von: http://www.avr-asm-tutorial.net/avr_de/rechnen/division.html
;*** dividiert 16 bit (div_h, div_l) durch 8 bit (divisor), liefert als Ergebnis
;*** einen 16-Bit-Wert (Register "quot_h", "quot_l")
dividiere:	
		clr		div_help
		ldi		quot_l, 1
		clr		quot_h

div_a:
		clc
		rol		div_l
		rol		div_h
		rol		div_help
		brcs	div_b
		cp		div_help, divisor
		brcs	div_c

div_b:
		sub		div_help, divisor
		sec
		rjmp	div_d
div_c:
		clc
div_d:
		rol		quot_l
		rol		quot_h
		brcc	div_a
		ret


;*** EEPROM einlesen, zu lesende Speicherstelle mu bei ****************
;*** Aufruf in temp stehen, Ergebnis dann auch in temp zu finden *******
read_eeprom:
		; EEPROM-Schreibvorgnge brauchen 3.4 ms zum beenden. Whrend dieser Zeit ist 
		; weder Schreiben noch Lesen auf dem EEPROM mglich, 
		; zur Kontrolle: EEPE = 1, whrend Schreibvorgang aktiv
		sbic	EECR, EEPE			; EEPE = 1? -> letzter Schreibvorgang nicht beendet
		rjmp	read_eeprom			; auf Beendigung warten

		out		EEAR, temp			; Addresse von EEPROM-Speicherstelle in EEAR laden
		sbi		EECR, EERE			; read-enable in EECR setzen
		in 		temp, EEDR			; Inhalt in temp lesen
		ret

;*** EEPROM schreiben, zu schreibende Speicherstelle mu bei ************
;*** Aufruf in temp stehen, zu schreibender Inhalt in temp2 *************
write_eeprom:
		; EEPROM-Schreibvorgnge brauchen 3.4 ms zum beenden. Whrend dieser Zeit ist 
		; weder Schreiben noch Lesen auf dem EEPROM mglich, 
		; zur Kontrolle: EEPE = 1, whrend Schreibvorgang aktiv
		delay 10
		sbic	EECR, EEPE			; EEPE = 1? -> letzter Schreibvorgang nicht beendet
		rjmp	write_eeprom		; auf Beendigung warten

		out		EEAR, temp			; Addresse von EEPROM-Speicherstelle in EEAR laden
		out		EEDR, temp2			; zu schreibenden Inhalt in EEDR laden
		sbi		EECR, EEMPE			; EEPROM_master_program_enable setzen
		sbi		EECR, EEPE			; Schreibvorgang durch setzen von EEPROM_program_enable starten
		ret

;********************************************************************************************
;********************** Programmparameter im EEPROM ablegen *********************************
;********************************************************************************************
save_params_in_eeprom:
	
		; programmparameter im EEPROM ablegen
		; Parameter jeweils in temp2 laden, und dann hintereinander weg ins EEPROM schreiben
		; Startaddresse: EEPROM-Speicherstelle 32
		; niedrigere EEPROM-Addresse fr Teachmode in Verwendung
		ldi		temp, 32
		ldi		temp2, ReedDelay
		rcall	write_eeprom

		inc		temp
		ldi		temp2, Delay_automatic
		rcall	write_eeprom

		inc		temp
		ldi		temp2, high(intervall)
		rcall	write_eeprom
		
		inc		temp
		ldi		temp2, low(intervall)
		rcall	write_eeprom

		inc		temp
		ldi		temp2, Hex_Offset
		rcall	write_eeprom

		inc		temp
		ldi		temp2, Rstrecke
		rcall	write_eeprom

		inc		temp
		ldi		temp2, TStrecke
		rcall	write_eeprom

		inc		temp
		ldi		temp2, Pumpstoesse
		rcall	write_eeprom

		inc		temp
		ldi		temp2, NoSignal
		rcall	write_eeprom

		inc		temp
		ldi		temp2, NotIntervall
		rcall	write_eeprom

		inc		temp
		ldi		temp2, PumpCross
		rcall	write_eeprom

		inc		temp
		ldi		temp2, WegCross
		rcall	write_eeprom

		inc		temp
		ldi		temp2, Pumpnormal
		rcall	write_eeprom

		inc		temp
		ldi		temp2, PumpRegen
		rcall	write_eeprom

		inc		temp
		ldi		temp2, WegRegen
		rcall	write_eeprom

		inc		temp
		ldi		temp2, RegenDelay
		rcall	write_eeprom

		inc		temp
		ldi		temp2, PumpTest
		rcall	write_eeprom

		inc		temp
		ldi		temp2, PumpOnTest
		rcall	write_eeprom

		inc		temp
		ldi		temp2, PumpOffTest
		rcall	write_eeprom

		inc		temp
		ldi		temp2, PumpOn
		rcall	write_eeprom

		inc		temp
		ldi		temp2, PumpOff
		rcall	write_eeprom

		inc		temp
		ldi		temp2, sleepmode
		rcall	write_eeprom

; on/off-optionen in 2 Byte zusammenfassen
; 1. Byte: Reeddelayon/off/Notlauf/Notlauf_dynamisch/Extralen(nach Regen)/Pumpimpulse im Regen/Speedprogression/CrossMode-Sofortpumpen
		ldi		temp2,0

#if Force_ReedDelay_On
		sbr		temp2,128
#endif

#if Force_ReedDelay_Off
		sbr		temp2,64
#endif
		
#if Enable_Notlauf
		sbr		temp2,32
#endif

#if Enable_Notlauf_Dynamisch
		sbr		temp2,16
#endif

#if Enable_ExtraOelen
		sbr		temp2,8
#endif

#if Add_Rain
		sbr		temp2,4
#endif

#if Enable_SpeedProgression
		sbr		temp2,2
#endif

#if Cross_SofortPumpen
		sbr		temp2,1
#endif

		ldi		temp, 64
		rcall	write_eeprom ; erstes Optionen-Byte an EEPROM (64) speichern

; 2. Byte: LED-Blinkereien
		ldi		temp2, 0

#if	Enable_BlinkNotlauf
		sbr		temp2,128
#endif

#if Enable_LED_Test
		sbr		temp2,64
#endif

#if	Enable_BlinkRain_for_Cross
		sbr		temp2,32
#endif

#if	Enable_Blink2sek_CrossLED
		sbr		temp2,16
#endif

#if	Enable_Blink15sek_CrossLED
		sbr		temp2,8
#endif

#if	Enable_Blink2sek_extRegenLED
		sbr		temp2,4
#endif

#if	Enable_Blink15sek_extRegenLED
		sbr		temp2,2
#endif

;platz fr ein weiteres Flag auf Bit0

		ldi		temp, 65
		rcall	write_eeprom ; zweites Optionen-Byte an EEPROM (65) speichern

		
; debug, counter fr pumpimpulse, max 255 -> genug km	
; initialisierung	
;		ldi		temp2, 0
;		ldi		temp, 66
;		rcall 	write_eeprom 
; end debug

; Progressionstabelle im EEPROM ablegen, Achtung: nur die ersten 16 Werte (sollten aber im Normalfall genug sein)
; Besonderes Format, um das EEPROM-File in z.B. Ponyprog besser menschenlesbar zu halten:
; Geschwindigkeit + zugehrige Progressionsstufe sind nicht direkt nebeneinander im EEPROM gespeichert, sondern um jeweils 16
; Speicherpltze getrennt
; damit ergibt es sich in PonyProg auch eine Tabellenform, so, das in der ersten Zeile der gespeicherten Tabelle
; die Geschwindigkeiten stehen, und in der Zeile direkt darunter die zugehrigen Progressionswerte
	
		ldi		ZH, high (Speedprogression*2)
		ldi		ZL, low (Speedprogression*2) ; Startaddresse der Tabelle laden
		ldi		temp, 96
save_speedtab:
		lpm		temp2, Z+			; ersten Wert (Geschwindigkeit) aus der Tabelle in temp2 laden
		rcall	write_eeprom		; im EEPROM ablegen

		subi	temp, -16			; Speicherstelle + 16
		lpm		temp2, Z+			; Progressionswert laden
		rcall	write_eeprom		; Im EEPROM ablegen

		subi	temp, 15			; Speicherstelle - 15 (also 1 neben der vorher gespeicherten Geschwindigkeit)
		
		cpi		temp, 112			; schon 16 Werte gespeichert?
		brsh	end_save_speedtab	; -> Ende

		ldi		temp2, low(end_speedprogression*2)
		cpse	Zl, temp2 ; Ende der Tabelle erreicht?

		rjmp	save_speedtab	; falls nein: Tabelle weiter abspeichern

end_save_speedtab:
		ret


;**************************************************************** 
; Timer0 Compare Match A        ; 
;**************************************************************** 
; Timer schafft ca. 15 Durchlufe pro Sekunde, daher z.B. der Faktor 15 fr Regendelay:
; konfigurierter Wert (z.b. 5) * 15 = 5*15 Timerdurchlufe, = 5 Sekunden Regendelay
; Andere Funktionen (Geschwindigkeitsberechnung, und damit einhergehend Wegverkrzungen) 
; arbeiten mit 2 Sekunden Intervall, daher weiter unten der Vergleichswert "31" fr die
; anzahl der Timerdurchlufe
; Einlesen des Hexschalters,
; Berechnen mglicher Intervallverkrzungen aufgrund:
; Geschwindigkeitsprogression, Regen, Crossmode
; die Wegverkrzungen erfolgen ber Manipulation (prozentuale Verringerung) der Radkonstante
; Regendelay

timer0_comp_match:

		push	temp				; Temp-Reg sichern
		push	temp2				
		in		temp,SREG			; Status-Reg sichern 
		push	temp

		sbis	PORTB,PB2			; richtige Sensorpolaritt prfen
		rjmp	Sensor_positiv
		cbi		PORTB,PB2			; ntigenfalls richtig setzen
		sbi		PORTB,PB3
		freq						; 0,25ms warten
Sensor_positiv:
; Regenstatus im FlagRegister speichern
		ldi		temp, 1
		sbic	ACSR,ACO			; Regen?
		add		cntRegenDelay, temp	; -> Regendelay-Zhler um 1 erhhen
		sbis	ACSR, ACO			; kein Regen
		sub		cntRegenDelay, temp ; ->Regendelay-Zhler um 1 verringern

		cp		cntRegenDelay, temp	; temp immer noch = 1 -> Vergleich Regenzhler mit 1
		brsh	kein_statuswechsel_trocken ; >=1? Status wird nicht auf trocken gesetzt
		cbi		Flagregister, Regen	; <1 -> status trocken abspeichern
		mov		cntRegenDelay, temp ; cntRegenDelay wieder auf 1 setzen, um einen Unterlauf 0 zu 255 zu verhindern
kein_statuswechsel_trocken:
		ldi		temp, cntRegenGrenze; Grenzwert fr Statuswechsel Regen laden
		cp		cntRegenDelay, temp ; Vergleich mit cntRegendelay, Grenzwert erreicht?
		brlo	kein_statuswechsel_nass ; Grenzwert nicht erreicht -> kein Statuswechsel
		sbi		FlagRegister, regen	; grenzwert erreicht -> statuswechsel Na durchfhren
		mov		cntRegenDelay, temp	; cntRegenDelay auf grenzwert setzen, um es nicht zu gro werden zu lassen

kein_statuswechsel_nass:
		inc		Cnt2sek				; Zhlregister erhhen
		ldi		temp, 31			; Vergleichswert fr 2-Sekunden-Intervall laden
		cpse	Cnt2sek, temp		; Timerintervall 31 mal abgelaufen = 2 Sekunden vergangen?
		rjmp	T0_CM_End			; nein: Routine beenden
		clr		Cnt2Sek				; Zhlreg. zurcksetzen

#if Enable_Blink2sek_CrossLED
		sbis	PortA, Cross_LED
		rjmp	no_blink2sek_cross
		cbi		PortA,Cross_LED
		delay	10
		sbi		PortA,Cross_LED
no_blink2sek_cross:
#endif
#if Enable_Blink2sek_extRegenLED
		sbis	PortD, ExternRegen
		rjmp	no_blink2sek_regen
		cbi		PortD,ExternRegen
		delay	10
		sbi		PortD,ExternRegen
no_blink2sek_regen:
#endif
		clr		temp				; Hexschalter einlesen
		sbis	PINB,4
		sbr		temp,1
		sbis	PINB,5				; Bit 1 und 2 des Hex-Schalters
		sbr		temp,4				; im Layout vertauscht
		sbis	PINB,6				; also zurcktauschen
		sbr		temp,2
		sbis	PINB,7
		sbr		temp,8

		subi	temp,-Hex_Offset	; Offset addieren (min 8 x 333m = 2666m)
		mov		wert,temp			; berechneten Vergleichswert im richtigen Reg. speichern
									; Umweg ber Temp ntig, da Wert ein Low-Register

		sbic	FlagRegister,Notbetrieb	; Notbetriebsmodus?
		rjmp	T0_CM_End			; dann Routine beenden

		ldi		temp, 2				; Inhalt von EEPROM-Speicherstelle 2 (=Radkonstante) laden
		rcall	read_eeprom			; und in SRAM speichern, um fr die mglicherweise gleich erfolgende
		sts		s_Rad, temp			; Verkrzung der Radkonstante (Regen, Crossmode) den richtigen Ausgangswert
									; im Speicher stehen zu haben. Wenn alleine mit der im SRAM gespeicherten
									; Radkonstante gearbeitet wrde, dann wird die bei jedem neuen Durchlauf immer
									; weiter verkrzt -> Fehler

#if Enable_SpeedProgression
		; Geschwindigkeitsprogression berechnen
		; dazu: erst die Geschwindigkeit aus der Anzahl der Radumdrehungen berechnen
		; danach die Progressionstabelle nach passendem Eintrag durchsuchen
		; und die Radkonstante mit dem gefundenen Wert anpassen

		; Berechnung der Geschwindigkeit:
		; Radumdrehung * 333/Radkonstante * 1,8
		; da ich aber nicht mit Zahlen > 255 multiplizieren kann, genausowenig mit gebrochen Zahlen
		; erfolgt die Berechnung im Programm etwas anders:
		; Radumdrehung * 167 (333:2)/Radkonstante * 36 (1,8*20) / 10
		lds		faktor_1, CntRad_prog		; Umdrehungsanzahl als ersten Faktor aus dem SRAM laden
		clr		temp2
		sts		CntRad_prog, temp2	; Umdrehungsanzahl im SRAM wieder auf 0 setzen
		ldi		faktor_2, halbes_intervall	; halbe Intervalllnge als 2. Faktor setzen
		rcall	multipliziere		
		; erster Teil: Umdrehung * 167 ist erledigt
		; Ergebnis der Multiplikation steht schon in den richtigen Registern, um gleich als Dividend
		; fr die folgende Division genutzt zu werden
		lds		divisor, s_Rad		; Radkonstante laden
		rcall	dividiere
		; zweiter Teil, Division durch Radkonstante ist erledigt
		; ergebnis der Division steht gleich im Faktor_1 - Register
		ldi		faktor_2, 36		; 36 als nchsten Faktor setzen
		rcall	multipliziere
		; dritter Teil: Multiplikation mit 36 erledigt
		ldi		divisor, 10			
		rcall	dividiere			; vierter Teil: Division durch 10 erledigt
		; Berechnung der Geschwindigkeit aus der Anzahl der Radumdrehungen abgeschlossen
		; noch zu prfen, ob Ergebnis kleiner 255, falls nicht, auf 255 setzen
		cpi		quot_h, 0			; enthlt das High-Byte des Divisionsergebnisses
									; irgendetwas <>0, war das Ergebnis grer als 255
		breq	speed_in_range		; falls =0 -> mit der Progressionsberechnung weitermachen
		ldi		quot_l, 255			; sonst auf 255 setzen
speed_in_range:
		mov		temp, quot_l		; Temp enthlt jetzt die aktuelle Geschwindigkeit

		; Geschwindigkeit berechnet, jetzt korrekten Wegfaktor
		; aus der "Speedprogression"-Tabelle suchen
		ldi		ZH, high (Speedprogression*2)
		ldi		ZL, low (Speedprogression*2) ; Startaddresse der Tabelle laden
		ldi		faktor_2, 100		; Der passende Verkrzungswert aus der Tabelle wird gleich ins entsprechende
									; Register fr die anschlieende Multiplikation geladen. 
									; Daher mit 100 (%) = keine Verkrzung initialisieren,
									; falls kein passender Wert in der Tabelle gefunden wird.
durchsuche_speedtab:
		lpm		temp2, Z+			; ersten Wert aus der Tabelle in temp2 laden
		cp		temp, temp2			; und mit berechneter Geschwindigkeit vergleichen
		brlo	berechne_speedprogression ; Geschwindigkeit (temp) < als Geschwindigkeit aus Tabelle (temp2)
									; dann zur Berechnung der Verkrzung springen
		cpi		ZL, Low(end_speedprogression*2) ; Ende der Tabelle erreicht?
		brsh	berechne_speedprogression	; Verkrzung berechnen

		lpm		faktor_2, Z+		; zur Geschwindigkeit gehrenden Prozentwert laden
									; die Geschwindigkeit fr diesen Wert ist kleiner als die aktuell gefahrene
									; wird beim nchsten Durchlauf dann eine grere Geschwindigkeit aus der Tabelle
									; gelesen, als gerade gefahren wird, dann haben wir den passenden Prozentwert gefunden
									; Bsp: Tabelle:
									; 100, 88
									; 130, 75
									; 160, 66
									; gefahren wird 140 km/h, richtiger Verkrzungswert ist also 75
									; 1. Durchlauf: - aus Tabelle gelesene Geschwindigkeit = 100 km/h
									;				- ist kleiner als aktuell gefahrene (140)
									;				-> Prozentwert (88) aus Tabelle laden
									;				-> Tabelle weiter durchsuchen
									; 2. Durchlauf:	- aus Tabelle gelesene Geschwindigkeit = 130 km/h
									;				- ist kleiner als aktuell gefahrene (140)
									;				-> Prozentwert (75) aus Tabelle laden
									;				-> Tabelle weiter durchsuchen
									; 3. Durchlauf:	- aus Tabelle gelesene Geschwindigkeit = 160 km/h
									;				- ist grer als aktuell gefahrene (140)
									;				-> zur Berechnung der Verkrzung springen, als Verkrzungswert
									;					wird die im 2. Durchlauf gelesene 75 verwendet

		rjmp	durchsuche_speedtab	; Tabelle weiter durchsuchen
berechne_speedprogression:
		lds		faktor_1, s_Rad		; Radkonstante als ersten Faktor laden
									; Faktor_2 ist ja schon passend belegt
 		rcall	multipliziere
		ldi		divisor, 100		; und noch die Division durch 100, wir wollen ja eine prozentuale Verkrzung erreichen
		rcall	dividiere
		cpi		quot_h, 0			; falls das High-Byte des Quotienten nach der Division <> 0 ist,
									; dann war "Radkonstante * Progressionsfaktor" > 255
									; damit kann nicht sinnvoll gearbeitet werden.
		breq	progression_in_range  ; falls =0, ist das Ergebnis hchstens 255 -> alles bestens, weitermachen
		ldi		quot_l, 255			; sonst auf 255 setzen -> maximale Intervallnge
progression_in_range:

		sts		s_Rad, quot_l		; angepate Radkonstante im SRAM ablegen
#endif

		sbis	Flagregister, Regen	; Regen?
		rjmp	NoRain				; Nein -> Sprung zu NoRain

		lds		faktor_1, s_Rad		; Regenverkrzung der Radkonstante berechnen
		ldi		faktor_2, WegRegen
		rcall	Multipliziere

		ldi		divisor, 100
		rcall	dividiere

		sts		s_Rad, quot_l		; neue, gekrzte Radkonstante im SRAM ablegen

NoRain:
		;sbic	PIND,Cross_SW		; PD0 Kontakt geschlossen? 
		sbis	Flagregister, Cross ; Crossmode?
		rjmp	T0_CM_End			; nein, dann keine Crossmode-Wegverkrzung
	
		lds		faktor_1, s_Rad		; Radkonstante
		ldi		faktor_2, WegCross	; Multiplikation vorbereiten, Faktoren laden
		rcall 	multipliziere		; Multiplikation durchfhren

		ldi		divisor, 100		; Divisor mit 100 laden
		rcall	dividiere			; Division durchfhren

		sts		s_rad, quot_l		; Low-Register des Ergebnisses bernehmen, High-Reg mu sowieso 0 sein

T0_CM_End:
		pop		temp
		out		SREG,temp			; Status Reg zurckschreiben 
		pop		temp2
		pop		temp				; Temp zurckschreiben
		reti

;**************************************************************** 
; Timer0 Compare Match B        ; 
; fr Reed-Entprellung benutzt	;
;**************************************************************** 
timer0_comp_match_B:

		push	temp				; Temp-Reg sichern
		in		temp,SREG			; Status-Reg sichern 
		push	temp
		
		ldi		temp, 255			; Compare_Match_Reg wieder auf 255 setzen
		out		OCR0B, temp			; um keine weiteren Interrupts zu erhalten
									; erst bei Aufruf des Reed_Interrupts wird dann
									; das Compare_Match_Register passend zum ReedDelay geladen

		ldi		temp, 0b01000000	
		out		GIMSK,temp			; ReedInterrupt wieder aktivieren
		out		EIFR,temp			; Interrupt-Flag zurcksetzen, da Interrupts innerhalb des Delays nur von
									; prellendem Sensor stammen knnen

		pop		temp
		out		SREG,temp			; Status Reg zurckschreiben 
		pop		temp				; Temp zurckschreiben
		reti

;**************************************************************** 
; Timer1 Compare Match         ; 
; fr das 15-Sekunden-Intervall (Basisintervall Notlauf)
;**************************************************************** 
timer1_comp_match:

		push	temp				; Temp-Reg sichern
		push	temp2
		in		temp,SREG			; Status-Reg sichern 
		push	temp

#if Enable_Blink15sek_CrossLED
		sbis	PortA, Cross_LED
		rjmp	no_blink15sek_cross
		cbi		PortA,Cross_LED
		delay	10
		sbi		PortA,Cross_LED
no_blink15sek_Cross:
#endif
#if Enable_Blink15sek_extRegenLED
		sbis	PortD, ExternRegen
		rjmp	no_blink15sek_regen
		cbi		PortD,ExternRegen
		delay	10
		sbi		PortD,ExternRegen
no_blink15sek_Regen:
#endif



		brts	T1_CM_End			; Testmode? -> Interrupt beenden

		inc		Timer1Cnt			; dann 15-Sek-Zhler erhhen

#if Enable_Notlauf
		sbis	PinA, Zuendung		; wenn Zuendung aus
		rjmp	T1_CM_End			; -> Interrupt beenden


		ldi		temp, NoSignal      ; Umweg ber Temp, da NoSignalReg ein Low-Register
		mov		NoSignalReg,temp	; NoSignal ins Register laden
		ldi		temp, NotIntervall
		mov		NotIntervallReg,temp; NotIntervall ins Register laden


#if Enable_Notlauf_Dynamisch
		; NotIntervall dynamisch berechnen

		; zuerst: Anpassung an Entfernungsintervall und 100 km/h
		; Register "wert" gibt an, wie oft das Entfernungsintervall zurckgelegt
		; werden mu, und enthlt damit quasi die Wegstrecke, die aktuell am McCoi eingestellt ist
		; daraus lt sich berechnen, welche Zeit vergehen mu, um diese Strecke
		; bei (ganz grob) 100 km/h zu bewltigen
		; Umrechnung erfolgt nach dem Schema:
		; wert * halbes_intervall / 208
		
		mov		faktor_1, wert
		ldi		faktor_2, halbes_intervall
		rcall	multipliziere

		ldi		divisor, 208
		rcall 	dividiere
		; das Ergebnis der Division kann (theoretisch) auch > 255 sein, aber praktisch glaube ich nicht,
		; das jemand den McCoi so extrem einstellt (es wren lentfernungen ber 100 km ntig, damit das passiert)
		; also spare ich mir hier den Vergleich
		mov		NotIntervallReg, quot_l ; angepates NotIntervall speichern

// Verkrzung des NotIntervalls fr Regen- und Crossmode berechnen
		sbis	Flagregister, Regen	; Regen?
		rjmp	NoRain_Notlauf		; Nein -> Sprung zu NoRain_Notlauf

		; Regenverkrzung des Notlauf-Intervalls berechnen
		; faktor_1 ist schon mit dem Ergebnis der ersten Berechnung passend belegt
		ldi		faktor_2, WegRegen
		rcall	Multipliziere

		ldi		divisor, 100
		rcall	dividiere

		mov		Notintervallreg, quot_l	; angepates Notintervall speichern

NoRain_Notlauf:
		;sbic	PIND,Cross_SW		; PD0 Kontakt geschlossen? 
		sbis	Flagregister, Cross ; Crossmode?
		rjmp	T1_CM_Weiter		; nein, dann keine Crossmode-Wegverkrzung
	
		; Crossmodeverkrzung des Notlauf-Intervalls berechnen
		; faktor_1 von vorherigen Berechnungen schon passend geladen
		ldi		faktor_2, WegCross	; Multiplikation vorbereiten, Faktoren laden
		rcall 	multipliziere		; Multiplikation durchfhren

		ldi		divisor, 100		; Divisor mit 100 laden
		rcall	dividiere			; Division durchfhren

		mov		Notintervallreg, quot_l		; Low-Register des Ergebnisses bernehmen, High-Reg mu sowieso 0 sein
#endif ; Ende berechnungen dynamischer Notlauf

T1_CM_Weiter:
		ldi		temp2, 1
		ldi		temp,0
		cp		temp, notintervallreg	; Notintervall auf = 0 berprfen
		in		temp, sreg				; sreg nach temp laden
		sbrc	temp, ze				; berprfen, ob Zero-Flag gesetzt
		add		notintervallreg, temp2  ; falls ja, dann temp2(=1) zu NotIntervallReg addieren
	
		sbic	FlagRegister,Notbetrieb	; Notbetrieb?
#if Enable_BlinkNotlauf
		rcall	blink_Regen				; wenn ja dann RegenLED blinken lassen, sonst normal weiter
#endif

		cp		Timer1Cnt, NotIntervallReg	; Notpumpintervall abgelaufen?
		brlo	T1_CM_End			; wenn nein -> Interrupt beenden

		sbic	FlagRegister,Notbetrieb	; Notbetrieb-Flag gesetzt?
		rjmp	Activate_Pump_T1

		cp		Timer1Cnt, NoSignalReg	; mehr Zeit als $NoSignal ohne Reedimpuls verstrichen?
		brlo	T1_CM_End			; nein -> Interrupt beenden
		sbi		FlagRegister, Notbetrieb	; ja -> Notbetrieb-Flag setzen

#if Enable_BlinkNotlauf
		rcall	blink_Regen
#endif
Activate_Pump_T1:
		clr		Timer1cnt
		ldi		Cr_soll,PumpNormal	; $PumpNormal laden
#if !Add_Rain
		sbic	FlagRegister, Regen ; Regen?
		ldi		Cr_soll, PumpRegen	; ja, und Add_Rain nicht  gesetzt -> PumpRegen laden
									; falls Add_Rain gesetzt erfolgt die Addition der Regenimpulse
									; erst am Ende, weil sie ja zu den Cross-Impulsen auch
									; dazugerechnet werden sollen
#endif
		;sbis	PIND,Cross_SW		; PD0 Kontakt geschlossen? 
		sbic	Flagregister, Cross ; Crossmode?
		ldi		Cr_soll,PumpCross	; ja, dann $PumpCross Impulse 
#if Add_Rain
		sbic	FlagRegister, Regen ; Regen?
		subi	Cr_soll, -PumpRegen	; ja, und Add_Rain gesetzt -> PumpRegen addieren
#endif
		;Pumpe aktivieren 
		clr		Cr_cnt				; damit startet die Pumpe im Hauptprogramm
#endif
T1_CM_End:
		pop		temp
		out		SREG,temp			; Status Reg zurckschreiben 
		pop		temp2
		pop		temp				; Temp zurckschreiben
		reti
;****************************************************************
; Initialisierung						;
;****************************************************************
Reset:	ldi		temp,low(ramend)	; initialize stackpointer <T.O.>
		out		spl,temp			;
	
		ldi		temp,0b00000000		; 
		out		MCUSR,temp			; Resetinformationen zurcksetzen

		out		WDTCR,temp			; Watchdogeinstellung vorbereiten
		ldi		temp,0b00001111		; 
		out		WDTCR,temp			; Watchdog 2s

		ldi		temp, 16
wait_startup:						; 16*250 ms = 4 sekunden verzgerung, um sauberes anklemmen 
									; an die Stromversorgung zu gewhrleisten
		delay	250		
		dec		temp
		brne	wait_startup


		ldi		temp,0b00111111
		out		PORTD,temp			; Pull-ups einschalten, LEDs aus
		ldi		temp,0b01110010
		out		DDRD,temp			; PD0 (CrossSW), PD2(Int0, Reed)
									; und PD3 (Taster) als Eingang
		sbi		PORTA,Cross_LED		; LED aus
		ldi		temp,0b00000001
		out		DDRA,temp			; Port A0 als Ausgang fr Cross-LED, PA1/PA2 als Inputs

		ldi		temp,0b11110000		; 
		out		PORTB,temp			; Port B Pull-ups im oberen Nibble ein
		ldi		temp,0b00001100		; Komparator und
		out		DDRB,temp			; oberes Nibble Port B als Eingang 


		ldi		temp,0b00110010		; Int 0 fallende Flanke, Powerdown
		out		MCUCR,temp 

		ldi		Cr_soll,0			; Standardwerte 0 Impulse
		ldi		temp,0x0F
		mov		Cr_cnt,temp			; Soviele wurde gepumpt
									; damits beim Reset nicht pumpt

		ldi		temp,0b01000000
		out		GIMSK,temp			; Int 0 Enable

		ldi		temp,252			; Compare Register A laden, 252*31*1024 CPU-Clocks = ca. 2 sekunden
		out		OCR0A,temp
		ldi		temp,255
		out		OCR0B,temp			; Compare-Register B fr Entprell-Delay vorbereiten

		ldi		temp,0b00000010		; Timer 0 konfigurieren
		out		TCCR0A,temp			; Prescaler 1024, CTC-Mode
		ldi		temp, 0b00000101	; weiteres Zhlregister fr T0-Compare-Matches, so da 
		out		TCCR0B,temp			; insgesamt 252*31*1024 CPU-Clocks = 2 Sekunden gezhlt werden knnen
			
		ldi		temp,CntTicksHigh	; CompareRegister laden (15 sek Timerlaufzeit), High-Wert
		out		OCR1AH,temp
		ldi		temp,CntTicksLow	; CompareRegister laden (15 sek Timerlaufzeit), Low-Wert
		out		OCR1AL,temp

		ldi		temp, 0b00000000	; Timer1 konfigurieren:
		out		TCCR1A, temp		; Prescaler 1024, Clear Timer on Compare Match-(CTC)-Mode
		ldi		temp, 0b00001101	; keine Compare-Outputs, da fr uns nur der
		out		TCCR1B, temp		; Compare-Match Interrupt interessant ist, um die
		ldi		temp, 0b00000000	; 15 Sekunden messen zu knnen
		out		TCCR1C, temp

		ldi		temp, 0b11000111	; TOIE0/OCIEA0/OCIEB0: Interrupt bei Timer0 Overflow/Compare Match
									; TOIE1/OCIEA1: Interrupt bei Timer1 Overflow/Compare Match
		out		TIMSK, temp			; Interrupt bei Compare Match zum Zhlen ermitteln der Intervalle

		clr		Timer1Cnt			; 15-Sek-Zhler auf 0 setzen
		sts		CntRad_prog,Timer1Cnt	; an 'CntRad_prog' eine 0 speichern

		ldi		temp,16
		mov		Wert,temp			; erstmal mit 5333m initialisieren (bis auslesen)

		; EEPROM testen, ob schon ordentliche Werte fr Tachopulse + Radkonstante gespeichert wurden
;		ldi		temp, 0				; Addresse von EEPROM-Speicherstelle 0 in temp laden		
;		rcall	read_eeprom			; und dann Inhalt von der Speicherstelle lesen
									; Ergebnis der Leseoperation auch in temp
;		cpi		temp, 0				; falls Inhalt der Speicherstelle ungleich 0,			
;		breq	init_SRAM_aus_EEPROM

		ldi		temp, 1
		rcall	read_eeprom			
		cpi		temp, 0xff			; und Speicherstelle 1 (Tachokonstante) = 0xff und
		brlo	init_SRAM_aus_EEPROM									
									
		ldi		temp, 2
		rcall	read_eeprom
		cpi		temp, 0xff			; und Speicherstelle 2 (Radkonstante) = 0xff
		brlo	init_SRAM_aus_EEPROM

		; mu EEPROM mit den Vorgabewerten fr Tacho- und Radkonstante neu beschrieben werden.
		; Annahme dabei: das EEPROM ist _standardmig_ komplett mit 0xff initialisiert, solange nichts
		; weiter hinein geschrieben wurde
;		ldi		temp, 0				; Addresse fr EEPROM-Schreibvorgang in temp laden (0 = EEPROM-Startaddresse)
;		ldi		temp2, 0			; Wert fr EEPROM-Schreibvorgang in temp2 laden
									; wird auf 0 gesetzt, als Fehlerkontrollmglichkeit
;		rcall	write_eeprom		

		ldi		temp, 1				; Addresse laden (1 = Tachokonstante)
		ldi		temp2, Tachopulse	; Wert laden (aus den Angaben am Programmanfang)
		rcall	write_eeprom

		ldi		temp, 2				; Addresse laden (2 = Radkonstante)
		ldi		temp2, Rad			; Wert laden
		rcall	write_eeprom
					
		rcall	save_params_in_eeprom ; konfigurierte Paramater im EEPROM ablegen


init_SRAM_aus_EEPROM:
		ldi		temp, 1				
		rcall	read_eeprom			; Inhalt von EEPROM-Speicherstelle 1 (=Tachokonstante) laden
		sts		s_Tachopulse, temp	; und in SRAM speichern


		ldi		temp, 2				
		rcall	read_eeprom			; Inhalt von EEPROM-Speicherstelle 2 (=Radkonstante) laden
		sts		s_Rad, temp			; und in SRAM speichern

		rcall	berechne_schwellwerte ; Schwellwerte "na" und "trocken" frs Nachlen berechnen und den Zhler initialisieren

		ldi		temp, 1
		mov		cntRegenDelay, temp ; Regendelayzhler initialisieren
		
		cbi		ACSR,ACD			; Komparator an

		clr		temp				; alle zuletzt verwendeten Register fr den Start des
		clr		temp2				; eigentlichen Programmes wieder zurcksetzen
		clr		c0
		clr		c1
		clr		c2
		clr		CntTacho			; Tachopulszhler zurcksetzen
		clr		cPump				; Pumpimpulszhler zurcksetzen
		clr		CntRad				; Umdrehungszhler lschen
		clr		CntWert				; 333m-Zhler zurcksetzen

		clr		XL
		clr		XH
#if Enable_LED_Test	
		rcall	LED_Test
#endif

		rjmp	Loop				; zum Anfang der Hauptschleife springen


Learnmode_Tacho:
		sbi		PortD, RegenLED		; LED zum Anzeigen der Lernmodes setzen
		cbi		PortD, ReedLed		; fr Tachopulse: Reed-LED an
		rjmp	Loop_end_Learnmode

Learnmode_Rad:
		cbi		PortD, RegenLED		; fr Radumfang: Regen-LED an
		sbi		PortD, ReedLed

Loop_end_Learnmode:
		sbi		PORTB,PB2			; Regensensor an
		cbi		PORTB,PB3
		freq
;****************************************************************
; Hauptschleife liest den Wegstrecken-Wert ein			;
;****************************************************************
Loop:	sei

		cbi		PORTB,PB2			; Regensensor an
		sbi		PORTB,PB3
		freq						; 0,25ms warten


;------------ abfrage Crossschalter/Taster
# if CrossMode_Taster
		sbic	PinD, Cross_SW		; Taster nicht gedrckt
		rjmp 	NoWait_CrossTaster	; keine Tasterauswertung
		ldi		temp, 3
		cp		cnt2sek, temp		; falls cnt2sek < 3, nix abziehen
		brlo 	noSub_Cnt2Sek
		sub		cnt2sek, temp		; cnt2sek um 3 verringern, um einen berlauf whrend des wartens auf den Taster zu verhindern.
noSub_Cnt2Sek:
		mov		temp, cnt2sek
		subi	temp, -2			; entprellzeit (2*durchlauf timer0)
Wait_CrossTaster:		
		wdr
		cp		temp, cnt2sek		; 2*Timer abgelaufen -> sollte genug entprellzeit gewesen sein
		breq	set_Crossmode		; dann zum Setzen des Crossmode springen
		sbis	PinD, Cross_SW		; ansonsten: solange Taste gedrckt
		rjmp	Wait_CrossTaster	; weiterwarten, bis 2 sekunden vorbei
		rjmp	NoWait_CrossTaster  ; Taster losgelassen -> zum ende springen
set_Crossmode:
		; toggle Bit in Flagreg
		clr		temp2
		sbr		temp2, (1<<cross)	; Bitmask fr xor vorbereiten, ermglicht toggle des Statusbits ohne vorherigen Status abfragen zu mssen
		in		temp, Flagregister
		eor		temp, temp2			
		out		Flagregister, temp	; Cross-status neu setzen
		
		; Cross-Status gleich anzeigen
		;Sbi		PORTA,Cross_LED		; Cross-LED aus
		;sbic		Flagregister, Cross	; Crossmode?
		;Cbi		PORTA,Cross_LED		; Cross-LED an

wait_release_crosstaster:
		wdr
		sbic	PinD, Cross_SW
		rjmp	NoWait_CrossTaster
		rjmp	wait_release_crosstaster	; warte auf Loslassen des Tasters
NoWait_CrossTaster:
# else
		cbi		Flagregister, Cross
		sbis	PIND, Cross_SW
		sbi		Flagregister, Cross
# endif
;-----------------------------------------


		in		temp,PinB			; Einlesen Hex-Schalter verkrzt,
		com		temp				; da nur Hex0 = Testmode an dieser
		andi	temp,0xF0			; Stelle festgestellt werden mu
		tst		temp

		brne	NoTest				; Testmode?

		set

		clr		Timer1Cnt			; 15 Sekunden-Zhler resetten

		in		temp, FlagRegister	; Inhalt FlagRegister in temp laden
		andi	temp, 0b00001111	; oberes Nibble (Lernmodi, Notlaufmodus) auf 0 setzen, damit diese abschalten
		out		FlagRegister, temp	; genderte Flags zurckschreiben

		sbi		PORTD,ReedLED		; LED aus
		sbic	PIND,ReedSW			; Reed kontakt offen?
		cbi		PORTD,ReedLED		; LED an

		Sbi		PORTA,Cross_LED		; Cross-LED aus
		;sbiS	PIND,Cross_SW		; Schalter geschlossen (on GND)?
		sbic	Flagregister, Cross	; Crossmode?
		Cbi		PORTA,Cross_LED		; Cross-LED an

#if Enable_BlinkRain_for_Cross
		;sbis	PIND,Cross_SW		; Schalter geschlossen (on GND)?
		sbic	Flagregister, Cross ; Crossmode?
		rcall 	blink_CrossMode
#endif

; Regenanzeige nach Status aus dem FlagRegister setzen
		sbi		PortD, RegenLED		; LED (intern) aus
		sbis	Flagregister, Regen	; kein Regen?
		cbi		PORTD,RegenLED		; LED (intern) an
		sbi		PortD, ExternRegen	; LED (extern) aus
		sbic	Flagregister, Regen	; Regen?
		cbi		PORTD,ExternRegen	; LED (extern) an

		ldi		temp,PumpTest		; Anzahl Pumpimpulse in Temp laden
		sbic	PIND,Taste			; Wenn Taste gedrckt
		mov		cPump,temp			; Anzahl Pumpimpulse von Temp nach cPump kopieren

Pump:	tst		cPump				; Anzahl Pulse abgelaufen?
		breq	jump_to_offset		; Ja, dann weiter
		sbic	PIND,Taste			; Taste gedrckt?
jump_to_offset:						; Hilfslabel, weil das Sprungziel fr das "breq Offset"
									; (2 zeilen weiter oben) auer Reichweite war
		rjmp	Offset				; Nein -> Weiter, sonst
		cli
		dec		cPump				; Pulszhler 
		sbi		PORTD,6				; Pumpe an
		delay	PumpOnTest
		cbi		PORTD,6				; Pumpe aus
		delay	PumpOffTest
		sei
		rjmp	Pump				; Weiter solange Taster gedrckt

NoTest:	clt							; Flag lschen
		; im folgenden der Abschnitt fr die Auswertung des Tasters im Betriebsmode
		; und das setzen des Learnmode
		sbic	PinD, Taste			; Taste gedrckt
		rjmp	End_Tastenauswertung; nein -> Tasterauswertung berspringen

		ldi		temp, 0				; Timer0 (2 Sekunden-Timer) reseten
		out		tcnt0, temp
		clr		cnt2sek

Wait_Taste:
		mov		temp, cnt2sek		
		cpi		temp, 30			; 30 * Timer abgelaufen -> knapp 2 Sekunden vorbei
		brsh	set_Learnmode		; dann zum Setzen des Learnmode springen
		sbis	PinD, Taste			; ansonsten: solange Taste gedrckt
		rjmp	Wait_Taste			; weiterwarten, bis 2 sekunden vorbei
		rjmp	End_Tastenauswertung; falls Taste vorher losgelassen -> beenden, keinen Learnmode setzen
set_Learnmode:		
		in		temp, FlagRegister
		cpi		temp, 0b01000000	; Bit 6 gesetzt? -> aktueller Zustand: Learnmode-Tacho
		brsh	enable_learn_rad	; zu Learnmode-Rad springen
enable_learn_tacho:
		sbi		FlagRegister, Learn_Tacho	; falls Bit 6 nicht gesetzt war -> jetzt setzen -> Learnmode-Tacho an
		ldi		XL, low (s_learn_tacho_start)	; Startaddresse fr Learnmode-Tacho laden
		sts		akt_speicherstelle_learnmode,XL ; und Learnmode-Speicherzeiger initialisieren
		clr		temp
clr_sram_learnmode:
		st		X+, temp
		cpi		XL, low(s_learn_rad_end)+1
		brlo	clr_sram_learnmode	; Speicherbereich fr Learnmode inititialisieren
		rjmp	End_Tastenauswertung
enable_learn_rad:
		cpi		temp, 0b10000000	; Bit 7 gesetzt? ja: aktueller Zustand: Learnmode-Rad
		brsh	disable_learnmodes	; falls ja: -> fertig mit lernen, Learnmode deaktivieren
									; falls nein: aktueller Zustand: Learnmode-Tacho
		;gelernte Anzahl der Tachopulse in EEPROM und SRAM speichern:
		ldi		XL, Low (s_learn_tacho_eeprom) ; X-Register (Speicherzeiger) initialisieren, Lowbyte
		ld		temp2, X			; temp2 (ins EEPROM zu schreibenden Wert) aus SRAM laden
		cpi		temp2, 1			; berprfen, ob mindestens 1 Tachopuls pro Radumdrehung ermittelt wurde
		brlo	nosave_Tacho		; falls nicht -> irgendein Fehler aufgetreten, keine Tachopulse speichern
		ldi		temp, 1				; EEPROM-Addresse in temp laden (1 = Tachopulse)
		rcall	write_eeprom		; Wert ins EEPROM schreiben
		sts		s_Tachopulse, temp2	; Wert im SRAM aktualisieren
nosave_tacho:
		cbi		FlagRegister, Learn_Tacho ; dann: Learnmode-Tacho deaktivieren
		sbi		FlagRegister, Learn_Rad	; Learnmode-Rad aktivieren
		ldi		XL, low (s_learn_rad_start)	; Startaddresse fr Learnmode-Rad laden
		sts		akt_speicherstelle_learnmode,XL ; und Learnmode-Speicherzeiger initialisieren
		rjmp	End_Tastenauswertung
disable_learnmodes:
		cbi		FlagRegister, Learn_Rad	; Learnmodes deaktivieren

; debug: alle whrend der Learnmodes ermittelten Werte nochmal im EEPROM ablegen, um sie hinterher 
; per Ponyprog auslesen und kontrollieren zu knnen
		ldi		XL, Low (s_learn_tacho_start) ; X-Register (Speicherzeiger) initialisieren, Lowbyte
		ldi		temp, 16			; Startaddresse EEPROM laden
debug_save_sram:
		ld		temp2, X+			; Speicher in Reg laden, postinkrement X-Pointer
		rcall	write_eeprom		; in EEPROM schreiben
		inc		temp				; Speicherstelle EEPROM eins weitersetzen
		cpi		XL, low (s_learn_rad_end)+1 ; Ende der Speicherstellen Learn-Mode erreicht?
		brlo	debug_save_sram		; wenn nein, weiterspeichern
; /debug

		; gelernte Daten fr Radkonstante in EEPROM und SRAM speichern
		ldi		XL, Low (s_learn_rad_eeprom) ; X-Register (Speicherzeiger) initialisieren, Lowbyte
		ld		faktor_1, X			; temp2 (ins EEPROM zu schreibenden Wert) aus SRAM laden
		cpi		faktor_1, 50			; berprfen, ob mindestens 50 Radumdrehungen pro 333 m ermittelt wurden
		brlo	End_Tastenauswertung; falls nicht -> irgendein Fehler aufgetreten, nix speichern
		; ermittelte Radkonstante (in faktor_1) basiert auf 333 Meter -> Umrechnen in anderes Basisintervall ntig
		; Rad * Intervall / 333 (bzw. Rad * Intervall/2 / 167
		ldi		faktor_2, halbes_intervall
		rcall	multipliziere
		ldi		divisor, 167
		rcall	dividiere
		mov		temp2, quot_l
		ldi		temp, 2				; EEPROM-Addresse in temp laden (2 = Tachopulse)
		rcall	write_eeprom		; Wert ins EEPROM schreiben
		sts		s_Rad, temp2		; Wert im SRAM aktualisieren
		rcall	berechne_Schwellwerte ; an die neue Radkonstante angepate Schwellwerte frs Nachlen berechen

End_Tastenauswertung:
		sbic	FlagRegister, Learn_Tacho
		rjmp	Learnmode_Tacho		
		sbic	FlagRegister, Learn_Rad
		rjmp	Learnmode_Rad		; im Learnmode nix weiter tun, wir springen direkt nach oben
									; zum setzen der LEDs auf der Platine
									; die sonstige Verabeitung (impulse Zhlen) erfolgt ausschlielich in der Reed-Routine
									 
		sbi		PORTD,ReedLED		; LED aus
		sbi		PORTD,RegenLED		; LED aus

		sbi		PORTD,ExternRegen	; LED aus an PD1
		sbic	FlagRegister, Regen	; Regen?
		cbi		PORTD,ExternRegen	; LED an an PD1

		/* Statusnderung Crossmode berwachen
		  Idee: - lade Temp mit 0
				- falls LED aus -> nix tun
				- falls LED an -> temp = temp+1
				- dann Crosschalter-Auswertung
				- falls LED an -> nix tun
				- falls LED aus -> temp = temp+1
				Ergebnis:	vorher aus, hinterher aus: temp = 1
							vorher aus, hinterher an: temp = 0
							vorher an, hinterher aus: temp = 1
							vorher an, hinterher aus: temp = 2
				- falls temp = 0 -> Zhlregister CntRad, CntWert setzen, um lung auszulsen*/

#if Cross_SofortPumpen
		ldi		temp, 0				
		sbis	PORTA, Cross_LED	; LED aus -> nchsten Befehl berspringen
		inc		temp				; LED an -> temp = temp + 1
#endif

		Sbi		PORTA,Cross_LED		; Cross-LED aus
		;sbis	PIND,Cross_SW		; Schalter geschlossen (on GND)?
		sbic	Flagregister, Cross ; Crossmode?
		Cbi		PORTA,Cross_LED		; Cross-LED an

#if	Cross_SofortPumpen
		sbic	PORTA, Cross_LED	; LED an -> nchsten Befehl berspringen
		inc		temp

		cpi		temp, 0				; temp = 0 -> lung auslsen?
		brne	kein_crossmodeanfang ; nein -> nix tun
		ldi		temp, rad			; Radkonstante laden
		mov		CntRad, temp		; Zhlregister Rad setzen
		mov		CntWert, Wert		; Zhlregister Wert setzen, beim nchsten Reed-Interrupt sollte es jetzt len
kein_crossmodeanfang:
#endif

#if Enable_BlinkRain_for_Cross
		;sbis	PIND,Cross_SW		; Schalter geschlossen (on GND)?
		sbic	Flagregister, Cross
		rcall 	blink_CrossMode
#endif

Offset:
		cp		Cr_cnt,Cr_soll		; Soll- und Istwert Pumpimpulse vergleichen
									; Falls _cnt < _soll so wird gepumpt (Werte werden in der Reed-Routine
									; entsprechend gesetzt), ansonsten wird das pumpen bersprungen
		brge	OffsPumpOK

AutoPump:
		cli
		sbi		PORTD,6				; Pumpe an
		delay	PumpOn				; warten
		cbi		PORTD,6				; Pumpe aus
		delay	PumpOff				; warten
		sei
		inc		Cr_cnt				; und 1 drauf
		cp		Cr_cnt,Cr_soll		; Fertig ?
		brlo	AutoPump			; Wenn kleiner, dann nochmal
		ldi		Cr_soll,0			; wieder auf Def 0 Impulse
;debug, counter fr pumpimpulse
;		ldi		temp, 66
;		rcall 	read_eeprom
;		inc		temp
;		mov		temp2, temp
;		ldi		temp, 66
;		rcall	write_eeprom
;end debug

OffsPumpOK:
		sbi		PORTB,PB2			; Regensensor negativ
		cbi		PORTB,PB3
		freq						; 0,25ms warten


#if Enable_Notlauf
		sbic	PinA,Zuendung		; Zndung an?
		rjmp	LoopEnd				; Wenn ja -> LoopEnd (nicht einschlafen lassen)
#endif

		mov		temp, Timer1Cnt
		cpi		temp, sleepmode
		brlo	LoopEnd				; wenn Timer1 2mal abgelaufen (=30 sek), dann
									; vorbereitungen zum einschlafen treffen
		sbic	FlagRegister,Notbetrieb	; im Notbetrieb?
		rjmp	Ausnahme			; dann einschlafen in der Ausnahme-Routine

		sbis	PIND,ReedSW			; wenn's Rad genau auf dem Reed steht
		rjmp	Ausnahme			; Sonderbehandlung

		sbi		PortA, cross_LED	; Cross-LED aus
		ldi		temp,0b00110110
		out		PORTD,temp			; Pull-ups fr Taster + Cross-SW
									; ausschalten, Pull-Ups fr LEDs (=LED aus)
									; und Reed bleiben aktiviert
		clr		temp
		out		PORTB,temp			; Port B Pull-ups aus
		sbi		ACSR,ACD			; Komparator aus
		ldi		temp,0b00011000		; 
		out		WDTCR,temp			; Watchdog aus vorbereiten
		clr		temp
		out		WDTCR,temp			; Watchdog aus
		ldi		temp,0b00110000		; Int 0 Level, Powerdown
		out		MCUCR,temp 

		sleep						; power down

		ldi		temp,0b00011000		; 
		out		WDTCR,temp			; Watchdog nderung vorbereiten
		ldi		temp,0b00001111		; 
		out		WDTCR,temp			; Watchdog an 2s
		ldi		temp,0b00110010		; Int 0 fallende Flanke, Powerdown
		out		MCUCR,temp
		cbi		ACSR,ACD			; Komparator an
		ldi		temp,0b11110000		; 
		out		PORTB,temp			; Port B Pull-ups im oberen Nibble ein
		ldi		temp,0b00111111
		out		PORTD,temp			; Port D Pull-ups einschalten

#if Enable_LED_Test
		rcall	LED_Test
#endif


LoopEnd:
		rjmp	Loop				; wieder von vorne beginnen

Ausnahme:
		clr		temp
		out		GIMSK,temp			; Int0/1 disable
		out		PORTB,temp			; Port B Pull-ups aus
		ldi		temp,0b00110110		; LEDs aus und Taste/Cross-SW auschalten
		out		PORTD,temp
		sbi		PortA,Cross_LED		; CrossLED aus
		sbi		ACSR,ACD			; Komparator aus
		
		wdr
		ldi		temp,0b00011000		; 
		out		WDTCR,temp			; Watchdognderung vorbereiten
		ldi		temp,0b01000111		; 
		out		WDTCR,temp			; Watchdog an 2s, Interruptmode

WD_Sleep:		; Sprungpunkt fr Watchdog-berwachtes Schlafen
		wdr

		sbic	FlagRegister,Notbetrieb	; im Notbetrieb?
		rjmp	Notlauf_Sleep		; dann Notlauf-Sleeproutine abarbeiten

		sbic	PIND,ReedSW			; Reed geschaltet
		rjmp	WakeUp				; Nein, dann aufwachen, sonst
		rjmp	Schlafen		

Notlauf_Sleep:

		ldi		temp,0b00110000		; Int 0 Level, Powerdown
		out		MCUCR,temp 
		ldi		temp,0b01000000
		out		GIMSK,temp			; Int 0 Enable

		sbic	PinA,Zuendung		; Zndung an?
		rjmp	WakeUp				; Ja -> aufwachen

Schlafen:
		sleep

		rjmp	WD_Sleep

WakeUp: 		; aufwachen und zurck zum regulren programmablauf
		ldi		temp,0b00110010		; Int 0 fallende Flanke, Powerdown
		out		MCUCR,temp
		ldi		temp,0b00111111		; PortD Pull-ups ein
		out		PORTD,temp
		cbi		ACSR,ACD			; Komparator ein
		ldi		temp,0b11110000		; 
		out		PORTB,temp			; Port B Pull-ups im oberen Nibble ein

		wdr
		ldi		temp,0b00011000		; 
		out		WDTCR,temp			; Watchdognderung vorbereiten
		ldi		temp,0b0001111		; 
		out		WDTCR,temp			; Watchdog an 2s, Resetmode

		ldi		temp,0b01000000
		out		GIMSK,temp			; Int 0 Enable

#if Enable_LED_Test
		rcall	LED_Test
#endif
		
		rjmp	Loop				; und wieder von vorne

;**************************************************************** 
; Reed Impuls, Zhlen         ; 
;**************************************************************** 
Reed:
		push	temp				; Temp-Reg sichern
		push	temp2
		in		temp,SREG			; Status-Reg sichern 
		push	temp

		clr		Timer1Cnt			; 15 Sekunden-Zhler resetten
		
		sbis	PIND,Taste			; Taste gedrckt? 
		rjmp	Reed_end

		brts	Reed_End_halfway	; Testmode -> Zurck

#if Force_ReedDelay_Off
		rjmp	no_reeddelay		; ReedDelay immer ausgeschaltet -> berspringen
#endif


		lds		temp, s_Tachopulse	; Anzahl Tachopulse laden
#if Force_ReedDelay_On
		ldi		temp, delay_automatic	; ReedDelay immer angeschaltet -> temp passend laden, so da der nachfolgende Vergleich
										; immer dazu fhrt, das Delay zu aktivieren
#endif
		cpi		temp, delay_automatic+1
		brsh	no_reeddelay		; Anzahl Tachopulse > Umschaltschwelle? -> kein Delay ntig, da vermutlich Tacho verwendet

		; ansonsten: Timer passend setzen, um den den Reedinterrupt fr die
		; gewnschte Zeit zu deaktivieren
		clr		temp
		out		GIMSK,temp			; Reed-Interrupt ausschalten

		in		temp,TCNT0			; aktuellen Zhlerstand (Counter0) laden
		ldi		temp2,EntprellTimer	; Anzahl Zhlerschritte fr Entprelldelay laden
		add		temp,temp2				; zu aktuellem Zhlerstand addieren
		brcs	Add_Overflow		; bichen Rechnerei, um Sonderflle: Overflow
		cpi		temp,252				; oder Wert grer als 252 abzufangen,
		brsh	ueber_252			; da der Timer0 ja nur bis 252 luft und nicht bis
		rjmp	Load_Entprelltimer	; zum berlauf bei 255
Add_Overflow:
		subi	temp, -3
		rjmp	Load_Entprelltimer
Ueber_252:
		subi	temp,252
Load_EntprellTimer:
		out		OCR0B,temp			; Timer0 Comp_Match_Reg_B mit Zhlerstand + Entprelldelay laden
									; beim Compare-Match (also nach Ablauf des Delays) wird dann der
									; Reed-Interrupt wieder eingeschaltet
no_reeddelay:
		cbi		FlagRegister,Notbetrieb	; kein Notbetrieb mehr

		sbic	FlagRegister,Learn_Tacho	; im Learnmode-Tachopulse? -> zur Speicherroutine springen
		rjmp	Lerne_Tachopulse

		inc		CntTacho   			; Zhler fr Tachopulse erhhen
		lds		temp,s_Tachopulse	; mit Vorgabewert fr Tachopulse vergleichen
		cp		CntTacho,temp		; wenn grer, weitermachen (umdrehungszhler erhhen...)
		brne	Reed_end_halfway	; sonst Reed-Routine beenden

		clr		CntTacho			; Zhler fr Tachopulse zurcksetzen

		sbic	FlagRegister,Learn_Rad	; im Learnmode-Radkonstante? -> zur Speicherroutine springen
		rjmp	Lerne_Radkonstante

		; bestimmen der Regenmenge in den letzten 333 m
		sbic	FlagRegister, Regen	; wenn Regen
		inc		Regenanteil			; dann Regenanteilszhler erhhen

		lds		temp, CntRad_prog	; Radumdrehungszhler fr Geschwindigkeitsprogression laden
		inc		temp				; um 1 erhhen
		sts		CntRad_prog, temp	; und wieder speichern
		rjmp	No_ReedEnd
Reed_End_halfway:					; Hilfs-Sprungmarke fr Reed_end
		rjmp	Reed_End
No_ReedEnd:

#if Enable_ExtraOelen
		lds		temp, CntRad_rain	; Zhler frs Nachlen laden
		dec		temp				; um 1 verringern
		breq	Nachoelen			; falls temp = 0 -> zum nachlen springen
		sts		CntRad_rain, temp	; sonst: Zhlerstand wieder speichern
		rjmp	Normal_Reed			; normal weitermachen



Nachoelen:
		rcall	berechne_Schwellwerte ; damit wird der Zhler frs Nachlen wieder passend gesetzt
		mov		temp2, Regenanteil	; da "Regenanteil" noch bentigt, gleich jedoch zurckgesetzt wird
									; den wert in "temp2" speichern
		clr		Regenanteil			; Regenanteil wieder auf 0 setzen, um die Regenmenge im nchsten 
									; Intervall zu bestimmen
;####################################################################
;########### Code zum len ---->NACH<---- Regen #####################

Regenpruefung:
		lds		temp, RSchwelle	
		cp		temp2,temp			; temp2 = Regenanteil, temp = Schwellwert na
									; letzte 333m berwiegend Regen?
									
		brlo	Trocken				; nein, auf berwiegend trocken prfen
		cpi		Regenmerker,RGrenze	; ja, deshalb Prfung auf mindestens 3000m Regen
		breq	Normal_Reed			; es hat (in der vergangenheit) mind. 3000m berwiegend
									; geregnet und regnet noch!
		inc		Regenmerker			; es regnet, aber noch keine 3000m Regenmerker erhhen!
		cpi		Regenmerker,RGrenze+1	; Hat es whrend dem Zusatzlen nach dem Regen
									; wieder angefangen zu Regnen?
		brlo	Normal_Reed			; nein, kein Eingriff ntig
		ldi		Regenmerker,RGrenze	; ja, Merker wird auf 3000m Regen gesetzt
		rjmp	Normal_Reed			; Programm normal weitermachen

Trocken:
		cpi		Regenmerker,0		; ist Regenmerker ungleich 0?
		breq	Normal_Reed			; gleich 0 (es IST und WAR berwiegend trocken)=> normal weitermachen
		cpi		Regenmerker,RGrenze	; ungleich 0 (es WAR berwiegend nass) => Hat es mindestens 3000m geregnet?
		brsh	Oelen				; ja, und es ist jetzt nicht mehr berwiegend nass (=> evtl. len)
		ldi		Regenmerker,0		; nein (max 2666m berwiegend Regen)
									; => Regenmerker wieder auf Trocken setzen!
		rjmp	Normal_Reed			; normal weiter, da nur kurz geregnet

Oelen:
		lds		temp, TSchwelle
		cp		temp2,temp			; temp2 = Regenanteil, temp = Schwellwert trocken
									; Ist es nach Regen schon abgetrocknet?
		brsh	Normal_Reed			; Regenanteil > TSchwelle -> noch nichts tun, da noch zu nass

		inc		Regenmerker			; Regenmerker wird auf 10-15 erhht!
									; (nach Regen ist es jetzt ziemlich trocken)
		cpi		Regenmerker,TGrenze	; 333m - 666m Trocken nach 1000m Regen?
		brlo	Normal_Reed			; ja, noch warten, [1000m oder mehr trocken => wird gelt]

		cpi		Regenmerker,PumpMax+1	; sind wir schon beim letzten. Mal Pumpen?
		brlo	Pumpen				; weniger als Pumpmax gepumpt: Pumpe aktivieren
		ldi		Regenmerker,0		; genug gepumpt: Regenmerker auf Trockenzustand setzen

;########### Code zum len ---->NACH<---- Regen #####################
;####################################################################
#endif
Normal_Reed:
		inc		CntRad				; Umdrehungszhler erhhen 
		lds		Temp,s_Rad
		cp		CntRad,temp			; Radkonstante erreicht? 
		brlo	Reed_end

		clr		CntRad				; Umdrehungszhler lschen 
		inc		CntWert				; 333m-Zhler erhhen 

		cp		CntWert,Wert		; Vergleichswert erreicht? 
		brlo	Reed_end 

Pumpen:
		ldi		Cr_soll,PumpNormal	; $PumpNormal laden
#if !Add_Rain
		sbic	FlagRegister, Regen ; Regen?
		ldi		Cr_soll, PumpRegen	; ja, und Add_Rain nicht  gesetzt -> PumpRegen laden
									; falls Add_Rain gesetzt erfolgt die Addition der Regenimpulse
									; erst am Ende, weil sie ja zu den Cross-Impulsen auch
									; dazugerechnet werden sollen
#endif
		;sbis	PIND,Cross_SW		; PD0 Kontakt geschlossen? 
		sbic	Flagregister, Cross ; Crossmode?
		ldi		Cr_soll,PumpCross	; ja, dann $PumpCross Impulse 
#if Add_Rain
		sbic	FlagRegister, Regen ; Regen?
		subi	Cr_soll, -PumpRegen	; ja, und Add_Rain gesetzt -> PumpRegen addieren
#endif

		clr		CntWert				; 333m-Zhler zurcksetzen 
		;Pumpe aktivieren 
		clr		Cr_cnt				; damit startet die Pumpe im Hauptprogr 
Reed_end:
		pop		temp
		out		SREG,temp			; Status Reg zurckschreiben 
		pop		temp2
		pop		temp				; Temp zurckschreiben
		reti						; Reed-Interrupt beenden

lerne_tachopulse:
		lds		XL, akt_speicherstelle_learnmode	; Speicherzeiger initialisieren
		ld		temp, X								; Inhalt von aktueller Speicherstelle laden
		inc		temp								; um eins erhhen
		brne	lerne_tachopulse_no_overflow		; falls nach dem Erhhen der Wert 0 ist,
													; trat ein Overflow aus
		ldi		temp, 255							; dann wieder den Wert auf 255 setzen
													; mehr als 255 Tachopulse pro Radumdrehungen geht halt nicht
lerne_tachopulse_no_overflow:
		st		X+, temp							; Wert wieder speichern, Speicherzeiger um 1 erhhen
		cpi		XL, Low (s_learn_tacho_end) + 1		; zeigt er noch auf eine Addresse innerhalb des gewnschten Bereichs?
		brlo	end_lerne_tachopulse				; ja -> beenden
		ldi		XL, low(s_learn_tacho_start)		; nein -> auf Anfang des Bereichs setzen
end_lerne_tachopulse:
		sts		akt_speicherstelle_learnmode, XL	; Speicherzeiger sichern
		rjmp	Reed_end							; Ende

lerne_radkonstante:
		lds		XL, akt_speicherstelle_learnmode	; Speicherzeiger initialisieren
		ld		temp, X								; Inhalt von aktueller Speicherstelle laden
		inc		temp								; um eins erhhen
		brne	lerne_radkonstante_no_overflow		; falls nach dem Erhhen der Wert 0 ist,
													; trat ein Overflow aus
		ldi		temp, 255							; dann wieder den Wert auf 255 setzen
													; mehr als 255 Radumdrehungen in 333 Metern geht halt nicht
lerne_radkonstante_no_overflow:
		st		X+, temp							; Wert wieder speichern, Speicherzeiger um 1 erhhen
		cpi		XL, Low (s_learn_rad_end) + 1		; zeigt er noch auf eine Addresse innerhalb des gewnschten Bereichs?
		brlo	end_lerne_radkonstante				; ja -> beenden
		ldi		XL, low(s_learn_rad_start)			; nein -> auf Anfang des Bereichs setzen
end_lerne_radkonstante:
		sts		akt_speicherstelle_learnmode, XL	; Speicherzeiger sichern
		rjmp	Reed_end							; Ende
;************************************** 
;################# PROGRAMM ENDE #########################
;#########################################################
;#########################################################
